/* SMAC_TEST.c                                            tab:4
 *
 * Copyright (c) 2001 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 * This application tests functionalitis of Sensor-MAC (S-MAC), and shows how
 * to use S-MAC. 
 *
 * Authors:	Wei Ye (USC/ISI)
 *
 */

#include "tos.h"
#include "include/smac_test_msg.h"
#include "SMAC_TEST.h"

#define NUM_TX_MSGS 10
#define NUM_FRAGS 10
#define PKT_INTERVAL 500

/* to test unicast, define UNICAST_ADDR by uncommenting the following 
 * line. Broadcast will be tested if UNICAST_ADDR is not defined.
 * Note: define my node address in MAIN.c
 *-------------------------------------------------------------------*/
//#define UNICAST_ADDR 1

#ifndef UNICAST_ADDR
#define MAC_TEST_BROADCAST
// change the number of neighbors according to your setup
#define NUM_NEIGHBORS 2
#define NUM_RX_MSGS (NUM_TX_MSGS * NUM_NEIGHBORS)
#endif

#ifdef UNICAST_ADDR
// following assumes only one neighbor will send to me
#define NUM_RX_MSGS NUM_TX_MSGS
#endif

#define TOS_FRAME_TYPE MAC_TEST_frame
TOS_FRAME_BEGIN(MAC_TEST_frame) {
	unsigned char numFrags;	// number of remaining fragments
	unsigned char numTx;	// number of Tx
	unsigned char numTxSucc; // number of successful Tx
	unsigned char numRx;	// number of received pkts
	unsigned char numRxSucc; // number of successful Rx
	unsigned char rxSeqNo;  // sequence no of received pkt
	unsigned char rxFragCount[NUM_RX_MSGS];
	unsigned int timeCount;	// when next packet will be generated
	AppPkt data; 		// message to be sent
	char send_pending;	// flag
}
TOS_FRAME_END(MAC_TEST_frame);


char TOS_COMMAND(TEST_INIT)(){
	VAR(numTx) = 0;
	VAR(numTxSucc) = 0;
	VAR(numRx) = 0;
	VAR(rxSeqNo) = NUM_RX_MSGS;
	VAR(send_pending) = 0;
	VAR(timeCount) = PKT_INTERVAL;
	TOS_CALL_COMMAND(TEST_SUB_INIT)(0);		// initialize lower components
	return 1;
}


char TOS_COMMAND(TEST_START)(){
	TOS_CALL_COMMAND(TEST_LEDy_on)();
	return 1;
}


void TOS_EVENT(TEST_SUB_SYNC_START)() {
}


TOS_TASK(sendMsg) {
	char success;
	// construct a new message
	VAR(data).hdr.seqNo = VAR(numTx);
	VAR(data).data[0] = VAR(numTxSucc);	 // no of successfully transmitted pkts
	VAR(data).data[1] = VAR(numRx);		 // no of received pkts so far
	VAR(data).data[2] = VAR(numRxSucc);	 // no of successfully received pkts

#ifdef MAC_TEST_BROADCAST
	success = TOS_CALL_COMMAND(TEST_SUB_BCAST_MSG)(&VAR(data), sizeof(VAR(data)));
#endif

#ifdef UNICAST_ADDR
	VAR(numFrags)= NUM_FRAGS;
	success = TOS_CALL_COMMAND(TEST_SUB_UCAST_MSG)(&VAR(data),
				sizeof(VAR(data)), UNICAST_ADDR, NUM_FRAGS);
#endif

	if (success) {
		VAR(send_pending) = 1;
	}
}


void TOS_EVENT(TEST_CLOCK_FIRE)() {
	if (VAR(timeCount) > 0) {
		VAR(timeCount)--;
		if (VAR(timeCount) == 0) {
			VAR(timeCount) = PKT_INTERVAL;
  			if (VAR(numTx) < NUM_TX_MSGS && VAR(send_pending) == 0) {
				TOS_POST_TASK(sendMsg);
			}
		}
	}
}


char TOS_EVENT(TEST_SUB_BCAST_MSG_DONE)(void* msg) {
#ifdef MAC_TEST_BROADCAST
	TOS_CALL_COMMAND(TEST_LEDr_tog)();
	VAR(numTx)++;
	if (VAR(numTx) == NUM_TX_MSGS) {
		// turn on LEDs to show final results
		TOS_CALL_COMMAND(TEST_LEDr_on)();
		if (VAR(numRx) == NUM_RX_MSGS) {
			TOS_CALL_COMMAND(TEST_LEDg_on)();
		}
	}		
	VAR(send_pending) = 0;
#endif
	return 1;
}


char TOS_EVENT(TEST_SUB_UCAST_FRAG_DONE)(void* frag) {
#ifdef UNICAST_ADDR
	TOS_CALL_COMMAND(TEST_LEDr_tog)();
	VAR(numFrags)--;
	if (VAR(numFrags) > 0) {
		TOS_CALL_COMMAND(TEST_SUB_SEND_MORE_FRAG)(&VAR(data));
	}
#endif // UNICAST_ADDR
	return 1;
}


void* TOS_EVENT(TEST_SUB_RX_MSG_DONE)(void* msg)
{
	AppPkt* pkt;
	pkt = (AppPkt*)msg;
	TOS_CALL_COMMAND(TEST_LEDg_tog)();
#ifdef MAC_TEST_BROADCAST
	if (((MACHeader*)msg)->toAddr == TOS_BCAST_ADDR) {
		VAR(numRx)++;
	}
#endif
#ifdef UNICAST_ADDR
	VAR(rxFragCount)[pkt->hdr.seqNo]++;
	if (VAR(rxFragCount)[pkt->hdr.seqNo] == NUM_FRAGS) {
		VAR(numRxSucc)++;
	}
	if (VAR(rxSeqNo) != pkt->hdr.seqNo) {
		VAR(numRx)++;
		VAR(rxSeqNo) = pkt->hdr.seqNo;
	}
#endif // UNICAST_ADDR
	if (VAR(numRx) == NUM_RX_MSGS) {
		// turn on LEDs to show final results
		TOS_CALL_COMMAND(TEST_LEDg_on)();
		if (VAR(numTx) == NUM_TX_MSGS) {
			TOS_CALL_COMMAND(TEST_LEDr_on)();
		}
	}		
	return msg;
}


char TOS_EVENT(TEST_SUB_UCAST_MSG_DONE)(unsigned char txFragCount){
	if (txFragCount > 0) {
		VAR(numTx)++;
		TOS_CALL_COMMAND(TEST_LEDr_tog)();
	}
	if (txFragCount == NUM_FRAGS) VAR(numTxSucc)++;
	if (VAR(numTx) == NUM_TX_MSGS) {
		// cancel timer
		VAR(timeCount) = 0;
		// turn on/off LEDs to show final results
		if (VAR(numTxSucc) == NUM_TX_MSGS) {
			TOS_CALL_COMMAND(TEST_LEDr_on)();
		} else {
			TOS_CALL_COMMAND(TEST_LEDr_off)();
		}
		if (VAR(numRx) == NUM_RX_MSGS) {
			TOS_CALL_COMMAND(TEST_LEDy_off)();
		}
	}
	VAR(send_pending) = 0;
	return 1;
}


void TOS_EVENT(TEST_SUB_SLEEP)()
{
	TOS_CALL_COMMAND(TEST_LEDy_off)();  // turn off yellow led when sleep
}


void TOS_EVENT(TEST_SUB_WAKEUP)()
{
	TOS_CALL_COMMAND(TEST_LEDy_on)();  // turn on yellow led when wake-up
}


