/*									tab:4
 * Copyright (c) 2002 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 * Authors:	Wei Ye
 * Date created: 7/21/2002
 *
 * The radio for Mica mote consists of 3 parts:
 *  1) RFM; 2) SPI; 3) Timer/counter 2
 *
 * This component controls the radio and provides the following funtions
 *  1) Put radio into 4 different states:
 *   	a) idle; b) sleep; c) receive; d) transmit
 *  2) Start symbol detection and bit synchronization
 *  3) Physical carrier sense (control is given to upper layer)
 *      Monitors the latest 16 bits. If the number of ones exceeds the
 *      BUSY_THRES, channel is determined as busy, and a signal is 
 *      generated. If channel keeps idle for the specified amount of time
 *      (in number of bits), a signal will be generated to indicate that
 *      the channel is idle.
 *
 */
 
#include "tos.h"
#include "dbg.h"
#include "RADIO_CONTROL.h"

//#define SP_DEBUG

// radio parameters
#define BIT_TIME 200
#define HALF_BIT_TIME (BIT_TIME / 2)
#define SAMPLE_TIME BIT_TIME
#define BUSY_THRES 5

#define MASK 0x7755775
#define STARTSYM 0x7044704
#define NUM_DISCARD 1

// radio state
#define SLEEP 0
#define IDLE 1
#define RECEIVE 2
#define TRANSMIT 3
#define INIT 4  /* temperary state only at start up */

// channel state
#ifdef CARR_SEN_TEST
#define CHAN_IDLE 0
#define CHAN_BUSY 1
#endif

static char start[5] = {0xcf, 0x0c, 0xcf, 0x0c, 0xcc};

#define TOS_FRAME_TYPE RADIO_CTRL_frame
TOS_FRAME_BEGIN(RADIO_CTRL_frame) {
	char state;
	unsigned long search;
	unsigned short carrSenTime; // carrier sense time
	unsigned short csBits;      // carrier sense bits
	unsigned char numOnes;      // carrier sense counter
	unsigned char numDiscard;   // number of bytes to be discarded in Rx
	char nextByte;  // tx buffer
	unsigned char txCount;   // for start symbol tx
#ifdef CARR_SEN_TEST
	char chanState;
#endif
}
TOS_FRAME_END(RADIO_CTRL_frame);


#ifdef SP_DEBUG
void ud_byte(char data)
{
  sbi(USR, TXC);
  outp(data, UDR); 
}

void ud_init()
{
  outp(12, UBRR);
  inp(UDR); 
  outp(0x08,UCR);
  sei();
}
#endif

	
char TOS_COMMAND(RADIO_INIT)() {
#ifdef SP_DEBUG
	ud_init();
#endif
#ifdef CARR_SEN_TEST
	VAR(chanState) = CHAN_IDLE;
#endif
	TOS_CALL_COMMAND(RADIO_SPI_SLAVE_INIT)();
	cbi(TIMSK, TOIE2); // disable timer 2 overflow interrupt
	VAR(state) = INIT; // just for changing to idle state
	TOS_CALL_COMMAND(RADIO_IDLE)();
	sei();  // enable globle interrupt
	
	return 1;
}

char TOS_COMMAND(RADIO_IDLE)() {
	if (VAR(state) == IDLE) return 1;
	if (VAR(state) != SLEEP) {
		// stop timer/counter 2
		outp(0x00, TCCR2); // stop SPI clock after tx/rx - important
		cbi(TIMSK, OCIE2); // disable compare match interrupt
		// turn off SPI
		outp(0x00, SPCR);
		TOS_CALL_COMMAND(RADIO_SPI_SLAVE_PIN_HIGH)();
	}
	//set RFM to Rx mode
	SET_RFM_CTL0_PIN();
	SET_RFM_CTL1_PIN();
	CLR_RFM_TXD_PIN();
	// clear state variables
	VAR(state) = IDLE;
	VAR(search) = 0;
	VAR(carrSenTime) = 0;
	VAR(csBits) = 0;
	VAR(numOnes) = 0;  // clear carrier sense counter
	VAR(numDiscard) = 0;
	// start timer/counter 2
	outp(0x09, TCCR2); // clear timer on compare match, no prescale
	outp(SAMPLE_TIME, OCR2); // set compare register
	outp(0x00, TCNT2); // clear current counter value
	sbi(TIMSK, OCIE2); // enable compare match interupt
	return 1;
}

char TOS_COMMAND(RADIO_SLEEP)() {
	if (VAR(state) == SLEEP) return 1;
	// stop timer/counter 2
	outp(0x00, TCCR2);
	cbi(TIMSK, OCIE2); // disable timer 2 compare match interrupt
	if (VAR(state) != IDLE) {
		// turn off SPI
		outp(0x00, SPCR);
		TOS_CALL_COMMAND(RADIO_SPI_SLAVE_PIN_HIGH)();
	}
	// set RFM to sleep mode
	CLR_RFM_TXD_PIN();
	CLR_RFM_CTL0_PIN();
	CLR_RFM_CTL1_PIN();
	VAR(state) = SLEEP;
	return 1;
}
	

/* this command starts a new packet transmission.
 * can only be called when radio is in idle or sleep state */

char TOS_COMMAND(RADIO_START_TX)() {
	char temp;
	outp(0x00, TCCR2);  // stop timer
	cbi(TIMSK, OCIE2);  // disable compare match interrupt
	if (VAR(state) != IDLE && VAR(state) != SLEEP) {
		// turn off SPI
		outp(0x00, SPCR);
		TOS_CALL_COMMAND(RADIO_SPI_SLAVE_PIN_HIGH)();
	}
	VAR(state) = TRANSMIT;
	VAR(nextByte) = start[1]; // buffer second byte of start symbol
	VAR(txCount) = 2;
	//set RFM to Tx mode
	CLR_RFM_CTL0_PIN();
	SET_RFM_CTL1_PIN();
	// start SPI
	temp = inp(SPSR);  // clear possible pending SPI interrupt
	outp(start[0], SPDR);  // put first byte into SPI data register
	outp(0xc0, SPCR);  // enable SPI and SPI interrupt
	TOS_CALL_COMMAND(RADIO_SPI_SLAVE_PIN_LOW)();
	//start timer/counter 2 to provide clock to SPI
	outp(0, TCNT2);
	outp(HALF_BIT_TIME, OCR2);
	sbi(DDRB, 7);   // set PB7 as output to provide clock signal to SPI
	cbi(PORTB, 7);  // set initial clock signal to low
	outp(0x19, TCCR2);  // toggle PB7 (OC2) on compare match
	return 1;
}	


/* This command spools next tx byte to SPI.
 * Can only be called after calling RADIO_START_TX, and
 * when radio signals RADIO_TX_BYTE_READY.
 * Make sure to turn radio into idle mode after finishing
 * each packet transmission/reception */
char TOS_COMMAND(RADIO_TX_NEXT_BYTE)(char data) {
	VAR(nextByte) = data;
	return 1;
}
	

char TOS_COMMAND(RADIO_START_CARR_SEN)(unsigned short numBits) {
	if (VAR(state) != IDLE) return 0;
	VAR(carrSenTime) = numBits;
	VAR(csBits) = 0;
	VAR(numOnes) = 0;
	return 1;
}


char TOS_EVENT(RADIO_SPI_SLAVE_NOTIFY)() {
    return 0;
}


/* Interrupt handler for SPI.
 * The signal handler disables globle interrupts by default. */
TOS_SIGNAL_HANDLER(SIG_SPI, ()){
	char data;
	data = inp(SPDR);
	if (VAR(state) == TRANSMIT) {
		outp(VAR(nextByte), SPDR);  // send buffered byte
		if (VAR(txCount) < sizeof(start)) {
			VAR(nextByte) = start[VAR(txCount)];
			VAR(txCount)++;
		} else {
			TOS_SIGNAL_EVENT(RADIO_TX_BYTE_READY)(); // ask a byte from upper layer
		}
	} else if (VAR(state) == RECEIVE) {
		outp(0, SPDR);
		if (VAR(numDiscard) < NUM_DISCARD) {
			VAR(numDiscard)++;
		} else {
			TOS_SIGNAL_EVENT(RADIO_RX_BYTE_DONE)(data);
#ifdef SP_DEBUG
//	ud_byte(data);
#endif
		}

	}		
}


/* Interrupt handler for timer/counter 2 compare match.
 * The signal handler disables globle interrupts by default.
 * This interrupt happens only when radio is in idle state. */

TOS_SIGNAL_HANDLER(SIG_OUTPUT_COMPARE2, ()){
	char bit;
	bit = READ_RFM_RXD_PIN() & 0x1;
	// search for start symbol
	VAR(search) <<= 1;
	VAR(search) |= bit;
	if ((VAR(search) & MASK) == STARTSYM){  // start symbol detected
		cbi(TIMSK, OCIE2); // disable compare match interrupt
		outp(0, TCCR2); // stop timer
		outp(HALF_BIT_TIME, OCR2); // for SPI receiving
		outp(35, TCNT2);  // may change the level of PB7
		sbi(DDRB, 7);  // set port B pin 7 as output
		cbi(PORTB, 7); // set port B pin 7 to high
		// clear possible pending SPI interrupt
		bit = inp(SPSR);
		outp(0, SPDR);
		// select SPI (ss low)
		TOS_CALL_COMMAND(RADIO_SPI_SLAVE_PIN_LOW)();
		// wait for a rising edge. need to read two 1s for noise rejection
 		bit = 0;
		do {
			if (READ_RFM_RXD_PIN()) bit++;
			else bit = 0;
		} while (bit < 2);
		// start counter and SPI to receive data
		// following line randomly changes level of PB7 on ATmega103,
		// but not on ATmega128
		outp(0x19, TCCR2); // toggle OC2 (PB7) on compare match
		VAR(state) = RECEIVE;
		// PINB can't be correctly read if immediately after starting timer
		if (inp(PINB) & 0x80) outp(0xc8, SPCR);
		else outp(0xc0, SPCR);
#ifdef SP_DEBUG
//	ud_byte(inp(PINB));
//	ud_byte(inp(TCNT2));
#endif
		// signal upper layer to prepare for reception
		TOS_SIGNAL_EVENT(RADIO_START_SYM_DETECTED)();
	} else {
		bit &= READ_RFM_RXD_PIN();
#ifndef CARR_SEN_TEST
		if (VAR(carrSenTime) > 0){  // carrier sense started
#endif
			VAR(numOnes) += bit;
			if ((VAR(csBits) & 0x8000) != 0) VAR(numOnes)--;
			VAR(csBits) = (VAR(csBits) << 1) | bit;

			if (VAR(numOnes) > BUSY_THRES) {
				// channel busy is detected
				VAR(carrSenTime) = 0;  // stop carrier sense
				TOS_SIGNAL_EVENT(RADIO_CHANNEL_BUSY)();
#ifdef CARR_SEN_TEST
				if (VAR(chanState) == CHAN_IDLE) {
					VAR(chanState) = CHAN_BUSY;
					TOS_SIGNAL_EVENT(RADIO_CHANNEL_STATE_CHANGE)(VAR(chanState));
				}
#endif
			} else {
				// channel idle is detected
				VAR(carrSenTime)--;
				if(VAR(carrSenTime) == 0){
					TOS_SIGNAL_EVENT(RADIO_CHANNEL_IDLE)();
				}
#ifdef CARR_SEN_TEST
				if (VAR(chanState) == CHAN_BUSY) {
					VAR(chanState) = CHAN_IDLE;
					TOS_SIGNAL_EVENT(RADIO_CHANNEL_STATE_CHANGE)(VAR(chanState));
				}
#endif
			}
#ifndef CARR_SEN_TEST
		}
#endif
	} 
	return;
}

