/* SMAC.c                                                      tab:4
 *
 * Copyright (c) 2001 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 *
 * Authors:  Wei Ye (USC/ISI)
 *
 * This module implements Sensor-MAC (S-MAC)
 * http://www.isi.edu/scadds/papers/smac_infocom.pdf
 *
 * It has the following functions.
 *  1) Low-duty-cycle operation on radio -- periodic listen and sleep
 *  2) Broadcast only uses CSMA
 *  3) Many features for unicast
 *     - RTS/CTS for hidden terminal problem
 *     - fragmentation support for a long message
 *       A long message is divided (by upper layer) into multiple fragments.
 *       The RTS/CTS reserves the medium for the entire message.
 *       ACK is used for each fragment for immediate error recovery.
 *     - Node goes to sleep when its neighbors are talking to other nodes.
 *
 */


extern short TOS_LOCAL_ADDRESS;

#include "tos.h"
#include "smac_msg.h"
#include "SMAC.h"

//#define SMAC_TEST
//#define SP_DEBUG
//#define SMAC_DEBUG
//#define MAC_TEST_SYNC
//#define FULLPC_SMAC_DEBUG

//#define PERFORMANCE
//#define MIN_NODE_ID 1

/* User-adjustable MAC parameters
 *--------------------------------
 * The default values can be overriden in Each application's Makefile
 * SMAC_MAX_NUM_NEIGHB: maximum number of neighbors.
 * SMAC_MAX_NUM_SCHED: maximum number of different schedules.
 * SMAC_DUTY_CYCLE: duty cycle in percentage. It controls the length of sleep 
 *   interval.
 * SMAC_RETRY_LIMIT: maximum number of RTS retries for sending a single message.
 * SMAC_EXTEND_LIMIT: maximum number of times to extend Tx time when ACK timeout
     happens.
 */
#ifndef SMAC_MAX_NUM_NEIGHB
#define SMAC_MAX_NUM_NEIGHB 20
#endif

#ifndef SMAC_MAX_NUM_SCHED
#define SMAC_MAX_NUM_SCHED 4
#endif

#ifndef SMAC_DUTY_CYCLE
#define SMAC_DUTY_CYCLE 10
#endif

#ifndef SMAC_RETRY_LIMIT
#define SMAC_RETRY_LIMIT 5
#endif

#ifndef SMAC_EXTEND_LIMIT
#define SMAC_EXTEND_LIMIT 5
#endif

/* Internal MAC parameters
 *--------------------------
 * Do NOT change them unless for tuning S-MAC
 * SLOTTIME: time of each slot in contention window, in ms. It should be large
 *   enough to receive the whole start symbol.
 * DIFS: DCF interframe space (from 802.11), in ms. It is used at the beginning
 *   of each contention window. It's the minmum time to wait to start a new 
 *   transmission.
 * SIFS: short interframe space (from 802.11), in ms. It is used before sending
 *   an CTS or ACK packet. It takes care of the processing delay of each pkt.
 * SYNC_CW: number of slots in the sync contention window, must be 2^n - 1 
 * DATA_CW: number of slots in the data contention window, must be 2^n - 1
 * GUARDTIME: guard time at the end of each listen interval, in ms.
 * SYNC_PERIOD: period to send a sync pkt, in ms.
 */
#define DIFS 10
#define SIFS 5
#define SLOTTIME 1
#define SYNC_CW 31
#define DATA_CW 63
#define GUARDTIME 4
#define SYNC_PERIOD 10000

/* Physical layer parameters
 *---------------------------
 * Based on the parameters from PHY_RADIO and RADIO_CONTROL
 * CLOCK_RES: clock resolution in ms. 
 * BANDWIDTH: bandwidth (bit rate) in kbps. Not directly used.
 * DIFS_BITS: number of bits in DIFS time. for carrier sense
 * BYTE_TX_TIME: time to transmit a byte, in ms. Derived from bandwidth
 * LISTEN_RATE: sampling rate of radio in listen mode (carrier sense)
 * PRE_PKT_BYTES: number of extra bytes transmitted before each pkt. It equals
 *   preamble + start symbol + sync bytes.
 * ENCODE_RATIO: output/input ratio of the number of bytes of the encoding
 *  scheme. In Manchester encoding, 1-byte input generates 2-byte output.
 * PROC_DELAY: processing delay of each packet in physical and MAC layer, in ms
 */
#define CLOCK_RES 1
#define BANDWIDTH 20
#define DIFS_BITS (DIFS * BANDWIDTH)
#define BYTE_TX_TIME 4/10  /* 0.4 ms to tx one byte */
#define LISTEN_RATE 20
#define PRE_PKT_BYTES 5
#define ENCODE_RATIO 2   /* Manchester encoding has 2x overhead */
#define PROC_DELAY 1

/* MAC states
 *-------------
 * SLEEP: radio is turned off, can't Tx or Rx
 * IDLE: radio in idle mode, will go to Rx if start symbol is deteded.
 *   Can start Tx in this state
 * CARR_SENSE: carrier sense. Do it before initiate a Tx
 * BACKOFF - medium is busy, and cannot Tx
 * WAIT_CTS - just sent RTS, and is waiting for CTS
 * WAIT_DATA - just sent CTS, and is waiting for DATA
 * WAIT_ACK - just sent DATA, and is waiting for ACK
 * TX_NEXT_FRAG - just send one fragment, waiting to tx next fragment
 */
#define SLEEP 0
#define IDLE 1
#define CHOOSE_SCHED 2
#define CARR_SENSE 3
#define BACKOFF 4
#define WAIT_CTS 5
#define WAIT_DATA 6
#define WAIT_ACK 7
#define TX_NEXT_FRAG 8
#define DATA_SENSE1 9
#define DATA_SENSE2 10
#define REPORT_RLT 11  /* for performance measurement */

// radio states
#define RADIO_SLEEP 0
#define RADIO_IDLE 1
#define RADIO_RX 2
#define RADIO_TX 3

// how to send a pkt: broadcast or unicast
#define BCAST_DATA 0
#define SEND_SYNC 1
#define SEND_RTS 2
#define SEND_CTS 3
#define SEND_DATA 4
#define SEND_ACK 5

// MAC packet types
#define DATA_PKT 0
#define RTS_PKT 1
#define CTS_PKT 2
#define ACK_PKT 3
#define SYNC_PKT 4
// followings are for debugging and performance measurement
#define START_PKT 5
#define STOP_PKT 6
#define RLT_TICKS_PKT 7
#define RLT_RX_PKT 8
#define RLT_TX_PKT 9
#define RLT_SLP_PKT 10
#define DEBUG_SLEEP 11
#define DEBUG_WAIT 12
#define DEBUG_DEST 13

// data type definitions
// note: pkt formats are defined in tos/include/smac_msg.h
typedef struct {
	char txSync;  // flag indicating need to send sync
	char txData;  // flag indicating need to send data
	unsigned short counter;  // tick counter
	unsigned short syncTime; // count for number of periods
} SchedTable;

typedef struct {
	short nodeId;
	unsigned char schedId;
} NeighbList;

#define TOS_FRAME_TYPE SMAC_frame
TOS_FRAME_BEGIN(MAC_SLEEP_frame) {
	// state variables
	char state;			// MAC state
	char opMode;        // mode of operation
	unsigned char numNeighb;  // number of known neighbors
	NeighbList neighbList[SMAC_MAX_NUM_NEIGHB]; // neighbor list
	unsigned char numSched;  // number of different schedules
	SchedTable schedTab[SMAC_MAX_NUM_SCHED];   // schedule table
	short mySyncNode; // who's my synchronizer
	char radioState;  // radio state
	
	// timing variables
	unsigned short syncTime;  // time of listen/contention interval for sync pkt
	unsigned short dataTime;  // time of listen/contention interval for data pkt
	unsigned short listenTime;  // time of listening interval
	unsigned short sleepTime;   // time of sleeping interval 
	unsigned short period;      // period of my schedule =listenTime + sleepTime
	unsigned char durDataPkt;   // duration (tx time needed) of data packet 
	unsigned char durCtrlPkt;   // duration (tx time needed) of control packet
	unsigned char durSyncPkt;   // duration (tx time needed) of sync packet
	unsigned char timeWaitCtrl; // time to wait for a control packet
	unsigned short nav;	// network allocation vector. nav>0 -> medium busy
	unsigned short neighbNav; // track neighbors' NAV while I'm sending/receiving
	unsigned short geneTime;    // generic timer
	unsigned char txDelay;	    // timer for carrier sense
	
	// Variables for Tx
	char txRequest;          // if I have accept a tx request;
	char howToSend;		     // what action to take for tx
	short sendAddr;		     // node that I'm sending data to
	unsigned char currSched; // current schedule I'm talking to
	unsigned char numRetry;	 // number of RTS tries for a data pkt
	unsigned char numExtend; // number of Tx time extensions when ACK timeout
	unsigned char numBcast;  // number of times needed to broadcast a packet
	unsigned char numFrags;  // number of fragments in this transmission
	unsigned char txFragCount; // number of transmitted fragments
	unsigned char txPktLen;  // length of data pkt to be transmitted
	MACHeader* dataPkt; // pointer to tx data pkt, only access MAC header
	MACCtrlPkt ctrlPkt;	  // MAC control packet
	MACSyncPkt syncPkt;   // MAC sync packet
	
	// Variables for Rx
	short recvAddr;		      // node that I'm receiving data from
	unsigned char lastRxFrag;  // fragment no of last received fragments

#ifdef SMAC_DEBUG
	char numLenErr;
	char numCrcErr;
	char numSlpPkt;
	char numCTStimeout;
	char numDATAtimeout;
	char numACKtimeout;
	char numSleeps;
	char numWakeups;
#endif
#ifdef PERFORMANCE
	unsigned long clockTime;    // clock time in mili-second
	unsigned long numSlpTicks;
	unsigned long numIdleTicks;
	unsigned long numRxTicks;
	unsigned long numTxTicks;
	RLT_Pkt rltPkt;		// outgoing result packet
#endif
}
TOS_FRAME_END(SMAC_frame);

// update my NAV
#define UPDATE_NAV(duration) do { \
	if (VAR(nav) < duration) VAR(nav) = duration; \
} while (0);

// track my neighbors' NAV
#define TRACK_NAV(duration) do { \
	if (VAR(neighbNav) < duration) VAR(neighbNav) = duration; \
} while (0);

#ifdef SP_DEBUG
void ud_byte(char data)
{
  sbi(USR, TXC);
  outp(data, UDR); 
}

void ud_init()
{
  outp(12, UBRR);
  inp(UDR); 
  outp(0x08,UCR);
  sei();
}
#endif

char TOS_COMMAND(MAC_INIT)(char opMode) {
#ifdef SP_DEBUG
	ud_init();
#endif
	// initialize state variables
	VAR(state) = CHOOSE_SCHED;
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
//	VAR(opMode) = LOWDUTYCYCLE;
	VAR(radioState) = RADIO_IDLE;
	VAR(numNeighb) = 0;  // number of known neighbors
	VAR(numSched) = 0;  // number of different schedules

	// initialize timing variables
	VAR(nav) = 0;
    VAR(neighbNav) = 0;
	VAR(txDelay) = 0;  // CHECK if we need
	
	// followings are constants that won't change once initialized
	VAR(durCtrlPkt) = (PRE_PKT_BYTES + sizeof(MACCtrlPkt) * ENCODE_RATIO)
						* BYTE_TX_TIME + 1;
/*
#ifdef SP_DEBUG
	ud_byte(VAR(durCtrlPkt));
#endif
*/
	VAR(durSyncPkt) = (PRE_PKT_BYTES + sizeof(MACSyncPkt) * ENCODE_RATIO)
						* BYTE_TX_TIME + 1;
	VAR(syncTime) = DIFS + SLOTTIME * SYNC_CW + VAR(durSyncPkt) + GUARDTIME;
	VAR(dataTime) = DIFS + SLOTTIME * DATA_CW + VAR(durCtrlPkt) + GUARDTIME;
	VAR(listenTime) = VAR(syncTime) + VAR(dataTime);
	VAR(period) = VAR(listenTime) * 100 / SMAC_DUTY_CYCLE;
/*
#ifdef SP_DEBUG
	ud_byte(VAR(period));
#endif
*/
	VAR(sleepTime) = VAR(period) - VAR(listenTime);
	// time to wait for CTS or ACK
	VAR(timeWaitCtrl) = PROC_DELAY + SIFS + VAR(durCtrlPkt) + PROC_DELAY;
/*
#ifdef SP_DEBUG
	ud_byte(VAR(timeWaitCtrl));
#endif
*/
	// listen for a whole period to choose a schedule first
	TOS_CALL_COMMAND(MAC_SUB_RAND_INIT)();
	VAR(geneTime) = SYNC_PERIOD + DIFS + 
			SLOTTIME * (TOS_CALL_COMMAND(MAC_SUB_NEXT_RAND)() & SYNC_CW);

	// initialize Tx variables
	VAR(txRequest) = 0;
	
	// fill in fixed portion of control packet
	VAR(ctrlPkt).fromAddr = TOS_LOCAL_ADDRESS;
	// fill in fixed portion of sync packet
	VAR(syncPkt).fromAddr = TOS_LOCAL_ADDRESS;
	VAR(syncPkt).type = SYNC_PKT;

#ifdef SMAC_DEBUG
    VAR(numLenErr) = 0;
    VAR(numCrcErr) = 0;
    VAR(numCTStimeout) = 0;
    VAR(numDATAtimeout) = 0;
    VAR(numACKtimeout) = 0;
    VAR(numSleeps) = 0;
    VAR(numWakeups) = 0;
#endif
#ifdef PERFORMANCE
	VAR(clockTime) = 0;
	VAR(numRxTicks) = 0;
	VAR(numTxTicks) = 0;
	VAR(numSlpTicks) = 0;
	VAR(rltPkt).fromAddr = TOS_LOCAL_ADDRESS;
#endif

    //initialize physical layer
    TOS_CALL_COMMAND(MAC_SUB_PHY_INIT)();
	// initialize and start clock
	TOS_CALL_COMMAND(MAC_SUB_CLOCK_INIT)(tick1000ps);  // 1 ms per tick
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC Module initialized\n");
#endif
    return 1;
}


char TOS_COMMAND(MAC_SEND_START_PKT)(char start) {
#ifdef PERFORMANCE
	// if this function is called, this node will send out a START_PKT,
	// and other node will reset their counter for performance measurement
	// this node will also become a synchronizer
	if (start) {  // send start packet
		VAR(ctrlPkt).type = START_PKT;
		VAR(ctrlPkt).addr = TOS_BCAST_ADDR;
		if (!TOS_CALL_COMMAND(MAC_SUB_TX_PACKET)((char*)&VAR(ctrlPkt)))
			return 0;
	} else {  // send stop packet
		VAR(state) = REPORT_RLT;
		VAR(rltPkt).type = STOP_PKT;
//		TOS_CALL_COMMAND(MAC_SUB_TIMERS_REGISTER)(TIMERPORT, 5);
	}
#endif
	return 1;
}


void setMySched(void* pkt) {
	// set my schedule and put it into the first entry of schedule table
	MACSyncPkt* packet = (MACSyncPkt*)pkt;
	VAR(state) = IDLE;
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
	VAR(numSched) = 1;
	VAR(schedTab)[0].syncTime = 0;
	VAR(schedTab)[0].txData = 0;
	VAR(schedTab)[0].txSync = 1;  // need to broadcast my schedule
	if (packet == 0) {  // freely choose my schedule
		VAR(schedTab)[0].counter = VAR(listenTime);  
		VAR(mySyncNode) = TOS_LOCAL_ADDRESS;
		// fill in the field in my syncPkt
		VAR(syncPkt).syncNode = VAR(mySyncNode); 
		VAR(currSched) = 0;  // broadcast my schedule to this schedule
		// schedule to send sync packet immediately
		sendSYNC();
    } else {  // follow the schedule in syncPkt
		VAR(schedTab)[0].counter = packet->sleepTime;
		VAR(mySyncNode) = packet->fromAddr;
		// fill in the field in my syncPkt
		VAR(syncPkt).syncNode = VAR(mySyncNode); 
		// add the node into my neighbor list
		VAR(neighbList)[0].nodeId = packet->fromAddr;
		VAR(neighbList)[0].schedId = 0;
		VAR(numNeighb) = 1;
    }
}

	
void TOS_COMMAND(MAC_POWER)(char mode) {
    if (mode == 0) {
		VAR(state) = SLEEP;
		VAR(radioState) = RADIO_SLEEP;
		TOS_CALL_COMMAND(MAC_SUB_RADIO_SLEEP)();  // turn off the radio
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
	} else {
		VAR(state) = IDLE;
		VAR(radioState) = IDLE;
		TOS_CALL_COMMAND(MAC_SUB_RADIO_IDLE)();  // turn on the radio
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
	}
}


void sleep() {
	// go to sleep, turn off radio and signal upper layer
	VAR(state) = SLEEP;
	TOS_CALL_COMMAND(MAC_SUB_RADIO_SLEEP)();  // turn off the radio
	VAR(radioState) = RADIO_SLEEP;
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
#ifdef SMAC_DEBUG
	VAR(numSleeps)++;
#endif
#ifdef SMAC_TEST
	TOS_SIGNAL_EVENT(MAC_SLEEP)();  // signal upper layer
#endif
}


void wakeup() {
	// wake up from sleep, turn on radio and signal upper layer
	VAR(state) = IDLE;
	TOS_CALL_COMMAND(MAC_SUB_RADIO_IDLE)();  // turn on the radio
	VAR(radioState) = RADIO_IDLE;
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
#ifdef SMAC_DEBUG
	VAR(numWakeups)++;
#endif
#ifdef SMAC_TEST
	TOS_SIGNAL_EVENT(MAC_WAKEUP)();  // signal upper layer
#endif
}


char TOS_COMMAND(MAC_BCAST_MSG)(void* data, unsigned char length) {
	unsigned char i;
	// Don't accept Tx request if I have already accepted a request
	// or have no neighbor information
	if (VAR(txRequest) != 0 || VAR(numNeighb) == 0 || data == 0)
		return 0;
	VAR(txRequest) = 1;
	VAR(dataPkt) = (MACHeader*)data;
	VAR(txPktLen) = length;
	VAR(sendAddr) = TOS_BCAST_ADDR;
	// fill in MAC header fields
	VAR(dataPkt)->type = DATA_PKT;  // data pkt
	VAR(dataPkt)->toAddr = TOS_BCAST_ADDR;
	VAR(dataPkt)->fromAddr = TOS_LOCAL_ADDRESS;
	VAR(dataPkt)->duration = 0;
	VAR(dataPkt)->fragNo = 0;
	// set flag in each schedule
	for (i = 0; i < VAR(numSched); i++) {
		VAR(schedTab)[i].txData = 1;
	}
	VAR(numBcast) = VAR(numSched);

	return 1;
}


char TOS_COMMAND(MAC_UCAST_MSG)(void* data, unsigned char length, short toAddr,
		unsigned char numFrags) {
	unsigned char i, foundNeighb, schedId = 0;
	// Don't accept Tx request if I have already accepted a request
	if (VAR(txRequest) != 0 || data == 0) return 0;
	foundNeighb = 0;
	for (i = 0; i < VAR(numNeighb); i++) {
		if (VAR(neighbList)[i].nodeId == toAddr) { // a known neighbor
			foundNeighb = 1;
			schedId = VAR(neighbList)[i].schedId;
			break;
		}
	}
	if (!foundNeighb) return 0;  // toAddr is unknow
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: got new unicast pkt to %x...\n", toAddr);
#endif
	VAR(txRequest) = 1;
   	VAR(dataPkt) = (MACHeader*)data;
	VAR(sendAddr) = toAddr;
	VAR(txPktLen) = length;
	VAR(numFrags) = numFrags;
	VAR(txFragCount) = 0;
	VAR(numRetry) = 0;
	VAR(numExtend) = 0;
	// calculate duration of data packet/fragment
	VAR(durDataPkt) = (PRE_PKT_BYTES + length * ENCODE_RATIO) 
					* BYTE_TX_TIME + 1;
	// fill in MAC header fields except duration
	VAR(dataPkt)->type = DATA_PKT;  // data pkt
	VAR(dataPkt)->toAddr = VAR(sendAddr);
	VAR(dataPkt)->fromAddr = TOS_LOCAL_ADDRESS;
	VAR(dataPkt)->fragNo = VAR(txFragCount);
	// set flag for data transmission
	VAR(schedTab)[schedId].txData = 1;
	return 1;
}

char TOS_COMMAND(MAC_TX_NEXT_FRAG)(void* data)
{
	// Send subsequent fragments
	if (VAR(state) != TX_NEXT_FRAG || data == 0) return 0;
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: got new fragment from upper layer.\n");
#endif
   	VAR(dataPkt) = (MACHeader*)data;
	// fill in MAC header fields except duration
	VAR(dataPkt)->type = DATA_PKT;  // data pkt
	VAR(dataPkt)->toAddr = VAR(sendAddr);
	VAR(dataPkt)->fromAddr = TOS_LOCAL_ADDRESS;
	VAR(dataPkt)->fragNo = VAR(txFragCount);
	// schedule to send this fragment
	VAR(howToSend) = SEND_DATA;
	VAR(txDelay) = SIFS;
	return 1;
}


void startBcast() {
	// broadcast data directly; don't use RTS/CTS
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: broadcasting packet...\n");
#endif
	TOS_CALL_COMMAND(MAC_SUB_TX_PACKET)(VAR(dataPkt), VAR(txPktLen));
	VAR(radioState) = RADIO_TX;
}


TOS_TASK(txMsgDone) {
	// prepare to tx next msg
	//VAR(state) = IDLE;  // CHECK if should sleep
	sleep();
	VAR(txRequest) = 0;
	TOS_SIGNAL_EVENT(MAC_UCAST_MSG_DONE)(VAR(txFragCount));
}


char TOS_EVENT(MAC_SUB_TX_PACKET_DONE)(void* packet) {
	char pktType;
	VAR(radioState) = RADIO_IDLE;
	if (packet == 0) return 0;
	pktType = *((char*)packet + sizeof(PhyHeader));
	switch (pktType) {  // the type field
	case SYNC_PKT:
		VAR(schedTab)[VAR(currSched)].txSync = 0;
		VAR(schedTab)[VAR(currSched)].syncTime = SYNC_PERIOD;
		break;
	case RTS_PKT:
		// track neighbors' NAV, need to clear it if CTS timeout
		TRACK_NAV(((MACCtrlPkt*)packet)->duration);
		// just sent RTS, set timer for CTS timeout
		VAR(geneTime) = VAR(timeWaitCtrl);
		break;
	case CTS_PKT:  // just sent CTS
		// track my neighbors' NAV
		// they update NAV and go to sleep after recv CTS
		TRACK_NAV(((MACCtrlPkt*)packet)->duration);
		// no data timeout, just use neighbors' NAV
		// since they went to sleep, just wait data for the entire time
		break;
	case DATA_PKT:
		if (VAR(howToSend) == BCAST_DATA) {
			// broadcast data is done for one schedule
//			VAR(state) = IDLE;
			sleep();
			VAR(schedTab)[VAR(currSched)].txData = 0;
			VAR(numBcast)--;
			if (VAR(numBcast) == 0) {
				// broadcast is done for all schedules
				VAR(txRequest) = 0;
				TOS_SIGNAL_EVENT(MAC_BCAST_MSG_DONE)(VAR(dataPkt));
			}
		} else {  // unicast is done
			// track my neighbors' NAV
			// they update NAV and go to sleep after recv first data fragment
			//TRACK_NAV(((MACHeader*)packet)->duration);
			// waiting for ACK, set timer for ACK timeout
			VAR(geneTime) = VAR(timeWaitCtrl);
		}
		break;
	case ACK_PKT:
		TRACK_NAV(((MACCtrlPkt*)packet)->duration); // in case tx extended.
		break;
	}
	return 1;
}


void TOS_EVENT(MAC_SUB_CLOCK_FIRE)() {
	// handle clock event
	unsigned char i, numRemain;
	unsigned short listenBits;
#ifdef PERFORMANCE
		// increment clock time
		VAR(clockTime)++;
		if (VAR(radioState) == RADIO_SLEEP)
			VAR(numSlpTicks)++;
		else if (VAR(radioState) == RADIO_IDLE)
			VAR(numIdleTicks)++;
		else if (VAR(radioState) == RADIO_RX)
			VAR(numRxTicks)++;
		else if (VAR(radioState) == RADIO_TX)
			VAR(numTxTicks)++;
#endif
	// NAV timer
	if (VAR(nav) > 0) {
		VAR(nav)--;
	}
	// timer to track my neighbor's NAV
	if (VAR(neighbNav) > 0) {
		VAR(neighbNav)--;
		if (VAR(neighbNav) == 0) {
			// used up reserved tx/rx time
//			if (VAR(state) == WAIT_DATA) {
			if (VAR(state) != SLEEP) {
				//VAR(state) = IDLE;  // CHECK if need to go to sleep
				sleep();
			}
		}
	}
	// generic timer
	if (VAR(geneTime) > 0) {
		VAR(geneTime)--;
		if (VAR(geneTime) == 0) {
			if (VAR(state) == CHOOSE_SCHED) {
				setMySched(0);  // I'm a primary synchronizer
			} else if (VAR(state) == DATA_SENSE1) {
				// didn't see a CTS reply to a previous RTS
				VAR(state) = DATA_SENSE2;
				VAR(geneTime) = VAR(timeWaitCtrl);
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
			} else if (VAR(state) == DATA_SENSE2) {
				// didn't see data tx, sender of RTS didn't get a CTS
				VAR(nav) = 0;  // cancel my NAV
				sleep();
			} else if (VAR(state) == WAIT_CTS) {	// CTS timeout
#ifdef SMAC_DEBUG
				VAR(numCTStimeout)++;
#endif
				VAR(neighbNav) = 0; // neighbors won't follow my reserved time
/*
#ifdef SP_DEBUG
	ud_byte(VAR(numCTStimeout));
#endif
*/
				if (VAR(numRetry) < SMAC_RETRY_LIMIT) {
					VAR(numRetry)++;
#ifdef FULLPC_SMAC_DEBUG
					printf("MAC: no CTS received. Goto backoff.\n");
#endif
					// wait unitl receiver's next wakeup
					//VAR(state) = IDLE;  // CHECK if need to go to sleep
					sleep();
				} else {		// reached retry limit, give up Tx
#ifdef FULLPC_SMAC_DEBUG
					printf("MAC: no CTS received. Reached retry limit, give up Tx!\n");
#endif
					// clear tx flag -- stop retry
					VAR(schedTab)[VAR(currSched)].txData = 0;
					TOS_POST_TASK(txMsgDone); // with txFragCount = 0;
				}
			} else if (VAR(state) == WAIT_ACK) {	// ACK timeout
#ifdef SMAC_DEBUG
				VAR(numACKtimeout)++;
#endif
				if (VAR(numExtend) < SMAC_EXTEND_LIMIT && 
					VAR(neighbNav) > VAR(durDataPkt)) {
					// can extend tx time
#ifdef FULLPC_SMAC_DEBUG
					printf("MAC: ACK timeout. Extend Tx time.\n");
#endif
					VAR(numExtend)++;
					// need to increase neighbNav for one frag
					numRemain = VAR(numFrags) - VAR(txFragCount);
					VAR(neighbNav) = numRemain * VAR(durCtrlPkt) + 
						numRemain * VAR(durDataPkt) +
						(numRemain + numRemain - 1) * (PROC_DELAY + SIFS);
					sendDATA();
				} else {
					// used up reserved time, stop Tx
					TOS_POST_TASK(txMsgDone);
				}
#ifdef PERFORMANCE
			} else if (VAR(state) == REPORT_RLT) {
				TOS_CALL_COMMAND(MAC_SUB_TX_PACKET)((char*)&VAR(rltPkt));
#endif
			}
		}
	}
	// update status of each schedule and arrange tx if needed
	for (i = 0; i < VAR(numSched); i++) {
		VAR(schedTab)[i].counter--;
		if (VAR(schedTab)[i].counter == VAR(listenTime)) {
			if (VAR(nav) == 0 && VAR(neighbNav) == 0 &&
				(VAR(state) == SLEEP || VAR(state) == IDLE)) {
				if (VAR(state) == SLEEP &&
					(i == 0 || VAR(schedTab)[i].txSync == 1)) {
					wakeup();  // wake up to listen or send sync
				}
				if (VAR(schedTab)[i].txSync == 1) {
					// start carrier sense for sending sync
					VAR(howToSend) = SEND_SYNC;
					VAR(currSched) = i;
					VAR(state) = CARR_SENSE;
					listenBits = (DIFS + SLOTTIME *
					(TOS_CALL_COMMAND(MAC_SUB_NEXT_RAND)() & SYNC_CW)) * LISTEN_RATE;
					TOS_CALL_COMMAND(MAC_SUB_START_CARR_SEN)(listenBits);
				}
			}
		} else if (VAR(schedTab)[i].counter == VAR(dataTime)) {
			if (VAR(schedTab)[i].txData == 1 &&
				VAR(nav) == 0 && VAR(neighbNav) == 0 &&
				(VAR(state) == SLEEP || VAR(state) == IDLE)) {
				// schedule sending of data
				if (VAR(state) == SLEEP) wakeup();  // wake up to tx
				if (VAR(sendAddr) == TOS_BCAST_ADDR)
					VAR(howToSend) = BCAST_DATA;
				else
					VAR(howToSend) = SEND_RTS;
				VAR(currSched) = i;
				VAR(state) = CARR_SENSE;
				listenBits = (DIFS + SLOTTIME *
				(TOS_CALL_COMMAND(MAC_SUB_NEXT_RAND)() & DATA_CW)) * LISTEN_RATE;
				TOS_CALL_COMMAND(MAC_SUB_START_CARR_SEN)(listenBits);
			}
		} else if (VAR(schedTab)[i].counter == 0) {
			VAR(schedTab)[i].counter = VAR(period);
			if (i == 0 && VAR(state) == IDLE && VAR(radioState) == RADIO_IDLE)
				sleep(); // sleep if I'm idle
		}
		// how long since my last sync
		if (VAR(schedTab)[i].syncTime > 0) {
			VAR(schedTab)[i].syncTime--;
			if (VAR(schedTab)[i].syncTime == 0) {
				VAR(schedTab)[i].txSync = 1;
			}
		}
	}

	// tx delay timer
	if (VAR(txDelay) > 0) {
		VAR(txDelay)--;
		if (VAR(txDelay) == 0) {  // start sending
			if (VAR(howToSend) == SEND_CTS) {
				sendCTS();
			} else if (VAR(howToSend) == SEND_DATA) {
				sendDATA();
			} else if (VAR(howToSend) == SEND_ACK) {
				sendACK();
			}
		}
	}
#ifdef SMAC_TEST
	// signal upper layer the clock event
	TOS_SIGNAL_EVENT(MAC_CLOCK_FIRE)();
#endif
	return;
}


char TOS_EVENT(MAC_SUB_CHANNEL_BUSY)(void) {
	// physical carrier sense indicate channel busy
	// Do nothing and stay in idle to receive a packet
	// Will sleep at my next sleep time if can't get a packet
	if (VAR(state) == CARR_SENSE) VAR(state) = IDLE;
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
	return 1;
}

char TOS_EVENT(MAC_SUB_CHANNEL_IDLE)(void) {
	// physical carrier sense indicate channel idle
	// start sending
	if (VAR(state) != CARR_SENSE) return 0;
	if (VAR(howToSend) == SEND_SYNC) {
		sendSYNC();
	} else if (VAR(howToSend) == BCAST_DATA) {
		startBcast();
	} else if (VAR(howToSend) == SEND_RTS) {
		sendRTS();
	}	
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif

	return 1;
}


char TOS_EVENT(MAC_SUB_START_SYM_DETECTED)() {
	VAR(radioState) = RADIO_RX;
	if (VAR(state) == IDLE || VAR(state) == CARR_SENSE) {	
		VAR(state) = BACKOFF;
	} else if (VAR(state) == DATA_SENSE2) {
		// data tx started from the sender of RTS
		sleep();
	}
	return 1;
}


void* TOS_EVENT(MAC_SUB_RX_PACKET_DONE)(void* packet, char error)
{
	char pktType;

	VAR(radioState) = RADIO_IDLE;
	if (VAR(state) == SLEEP) {
		return packet; // if in sleep, reject any pkt
#ifdef SMAC_DEBUG
		VAR(numSlpPkt)++;
#endif
	}

	if (error) {  // if received an erroneous pkt, a sign of collision
#ifdef SMAC_DEBUG
		if (packet == NULL) VAR(numLenErr)++;
		else VAR(numCrcErr)++;
#endif
		if (VAR(state) == BACKOFF) sleep();  // wait until next tx time
		return packet;
	}
	
	pktType = *((char*)packet + sizeof(PhyHeader));
	// dispatch to actual packet handlers
	if (pktType == DATA_PKT) {
		return handleDATA(packet);
	} else if (pktType == RTS_PKT) {
		handleRTS(packet);
	} else if (pktType == CTS_PKT) {
		handleCTS(packet);
	} else if (pktType == ACK_PKT) {
		handleACK(packet);
	} else if (pktType == SYNC_PKT) {
		handleSYNC(packet);
	}
	return packet;
}


void handleRTS(void* pkt) {
	// internal handler for RTS
	MACCtrlPkt* packet;
	if (VAR(state) != IDLE && VAR(state) != BACKOFF) return;
	packet = (MACCtrlPkt*)pkt;
	if (packet->toAddr == TOS_LOCAL_ADDRESS) {
		VAR(recvAddr) = packet->fromAddr;  // remember sender's address
		VAR(lastRxFrag) = 250;
#ifdef FULLPC_SMAC_DEBUG
		printf("MAC: Received RTS from: %x\n", VAR(recvAddr));
#endif
		TRACK_NAV(packet->duration); // track neighbors' NAV
		// schedule sending CTS
		VAR(howToSend) = SEND_CTS;
		VAR(txDelay) = SIFS;
	} else { // packet destined to another node
		// assume sender will get a CTS and tx data
		UPDATE_NAV(packet->duration);
		// keep listening until confirm sender gets a CTS or starts tx data
		VAR(state) = DATA_SENSE1; // wait for a CTS
		VAR(geneTime) = VAR(timeWaitCtrl);
#ifdef SP_DEBUG
		ud_byte(VAR(state));
#endif
	}
}
		

void handleCTS(void* pkt) {
	// internal handler for CTS
	MACCtrlPkt* packet;
	packet = (MACCtrlPkt*)pkt;
	if (packet->toAddr == TOS_LOCAL_ADDRESS) {
		if (VAR(state) == WAIT_CTS && packet->fromAddr == VAR(sendAddr)) {
			// cancel CTS timer
			VAR(geneTime) = 0;
#ifdef FULLPC_SMAC_DEBUG
			printf("MAC: Received CTS from: %x\n", VAR(sendAddr));
#endif
			VAR(schedTab)[VAR(currSched)].txData = 0;
			// schedule sending DATA
			VAR(howToSend) = SEND_DATA;
			VAR(txDelay) = SIFS;
		}
	} else { // packet destined to another node
		if (VAR(state) == DATA_SENSE1 || VAR(state) == DATA_SENSE2) {
			VAR(geneTime) = 0;
		}
		if (VAR(state) == IDLE || VAR(state) == BACKOFF || 
			VAR(state) == DATA_SENSE1 || VAR(state) == DATA_SENSE2) {
			UPDATE_NAV(packet->duration);
			sleep();
		}
	}
}


void* handleDATA(void* pkt) {
	// internal handler for DATA packet
	void* tmp;
	MACHeader* packet;
	packet = (MACHeader*)pkt;
	if (packet->toAddr == TOS_BCAST_ADDR) {  // broadcast packet
#ifdef FULLPC_SMAC_DEBUG
		printf("MAC: Received broadcast data pkt from: %x\n", packet->fromAddr);
#endif
//		VAR(state) = IDLE;		// CHECK if should sleep
		sleep();
		tmp = TOS_SIGNAL_EVENT(MAC_RX_MSG_DONE)(packet);
		return tmp;
	} else if (packet->toAddr == TOS_LOCAL_ADDRESS) {  // unicast packet
		// could receive data in idle, backoff and wait_data state
		if (VAR(state) == WAIT_DATA && packet->fromAddr == VAR(recvAddr)) {
#ifdef FULLPC_SMAC_DEBUG
			printf("MAC: Received unicast data pkt from: %x\n", packet->fromAddr);
#endif
			// Should track neighbors' NAV, in case tx extended
			TRACK_NAV(packet->duration);
			// schedule sending ACK
			VAR(howToSend) = SEND_ACK;
			VAR(txDelay) = SIFS;
			// if fragNo is the same as last one, don't signal duplicate
			if (VAR(lastRxFrag) != packet->fragNo) {
				VAR(lastRxFrag) = packet->fragNo;
				tmp = TOS_SIGNAL_EVENT(MAC_RX_MSG_DONE)(packet);
				return tmp;
			}
		}
	} else { // unicast packet destined to another node
		if (VAR(state) == IDLE || VAR(state) == BACKOFF) {
			UPDATE_NAV(packet->duration);
			sleep();
		}
	}
	
	return pkt;
}


void handleACK(void* pkt) {
	// internal handler for ACK packet
	MACCtrlPkt* packet;
	packet = (MACCtrlPkt*)pkt;
	if (packet->toAddr == TOS_LOCAL_ADDRESS) {
		if (VAR(state) == WAIT_ACK && packet->fromAddr == VAR(sendAddr)) {
			// cancel ACK timer
			VAR(geneTime) = 0;
#ifdef FULLPC_SMAC_DEBUG
			printf("MAC: Received ACK from: %x\n", VAR(sendAddr));
#endif
			VAR(txFragCount)++;  // number of successfully transmitted frags

			if (VAR(txFragCount) < VAR(numFrags)) {
				if (VAR(neighbNav) >= (SIFS + VAR(durDataPkt) + 
					VAR(timeWaitCtrl))) {
					// have enough time for next fragment
					// ask upper layer for next fragment
					VAR(state) = TX_NEXT_FRAG;
					TOS_SIGNAL_EVENT(MAC_UCAST_FRAG_DONE)(VAR(dataPkt));
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
				} else {  // no more time for tx
					VAR(state) = IDLE; // won't accept next frag
					TOS_SIGNAL_EVENT(MAC_UCAST_FRAG_DONE)(VAR(dataPkt));
					TOS_POST_TASK(txMsgDone);
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
				}
			} else { // no more fragment, tx done
				VAR(state) = IDLE;
				TOS_SIGNAL_EVENT(MAC_UCAST_FRAG_DONE)(VAR(dataPkt));
				TOS_POST_TASK(txMsgDone);
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
			}
		}
	} else { // packet destined to another node
		if (VAR(state) == IDLE || VAR(state) == BACKOFF) {
			UPDATE_NAV(packet->duration);
			sleep();
		}
	}
}


void handleSYNC(void* pkt)
{
	// internal handler for SYNC packet
	MACSyncPkt* packet;
	unsigned char i, foundSched, foundNeighb;
	unsigned char schedId;
	packet = (MACSyncPkt*)pkt;
	if (VAR(state) == CHOOSE_SCHED) {
		VAR(geneTime) = 0;  // cancel timer
		setMySched(packet);
		return;
	}
	// following is excuted after initial choosing of schedule
	// if cs failed and received a sync pkt, go back to idle
	VAR(state) = IDLE;
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
	// check if sender is on my neighbor list
	foundNeighb = 0;
	schedId = SMAC_MAX_NUM_SCHED;
	for (i = 0; i < VAR(numNeighb); i++) {
		if (VAR(neighbList)[i].nodeId == packet->fromAddr) {
			foundNeighb = 1;
			schedId = VAR(neighbList)[i].schedId; // a known neighbor
			// update corresponding schedule
			VAR(schedTab)[schedId].counter = packet->sleepTime;
			break;
		}
		if (VAR(neighbList)[i].nodeId == packet->syncNode)
			// found its synchronizer, remember it schedule id
			schedId = VAR(neighbList)[i].schedId;
	}
	if (!foundNeighb) {  // unknown node, add it onto neighbor list
		VAR(neighbList)[VAR(numNeighb)].nodeId = packet->fromAddr;
		if (schedId < SMAC_MAX_NUM_SCHED) {
			// found its synchronizer
			VAR(neighbList)[VAR(numNeighb)].schedId = schedId;
		} else if (packet->syncNode == TOS_LOCAL_ADDRESS) {
			// This node follows my schedule
			VAR(neighbList)[VAR(numNeighb)].schedId = 0;
		} else {  // its synchronizer is unknown
			// check if its schedule equals to an existing one
			foundSched = 0;
			for (i = 0; i < VAR(numSched); i++) {
				if (VAR(schedTab)[i].counter == packet->sleepTime ||
					VAR(schedTab)[i].counter == packet->sleepTime + 1 ||
					VAR(schedTab)[i].counter + 1 == packet->sleepTime) {
					VAR(neighbList)[VAR(numNeighb)].schedId = i;
					foundSched = 1;
					break;
				}
			}
			if (!foundSched) {  // this is an unknown schedule
				VAR(schedTab)[VAR(numSched)].counter = packet->sleepTime;
				VAR(schedTab)[VAR(numSched)].txSync = 1;  // need send sync
				VAR(schedTab)[VAR(numSched)].txData = 0;
				VAR(neighbList)[VAR(numNeighb)].schedId = VAR(numSched);
				VAR(numSched)++;  // increment number of schedules
			}
		}
		VAR(numNeighb)++;  // increment number of neighbors
	}
/*
#ifdef SP_DEBUG
	ud_byte(VAR(numNeighb));
#endif
*/
}


void sendRTS() {
	// construct and send RTS packet
	VAR(ctrlPkt).type = RTS_PKT;
	VAR(ctrlPkt).toAddr = VAR(sendAddr);
	// reserve time for CTS + all fragments + all ACKs
	VAR(ctrlPkt).duration = (VAR(numFrags) + 1) * VAR(durCtrlPkt) + 
					VAR(numFrags) * VAR(durDataPkt) +
					(VAR(numFrags) + VAR(numFrags) + 1) * (PROC_DELAY + SIFS);
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: Sending RTS to: %x\n", VAR(sendAddr));
#endif
	// send RTS
	TOS_CALL_COMMAND(MAC_SUB_TX_PACKET)(&VAR(ctrlPkt), sizeof(VAR(ctrlPkt)));
	VAR(state) = WAIT_CTS;
	VAR(radioState) = RADIO_TX;
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
}


void sendCTS() {
	// construct and send CTS
	// input duration is the duration field from received RTS pkt
	VAR(ctrlPkt).type = CTS_PKT;
	VAR(ctrlPkt).toAddr = VAR(recvAddr);
	// should track neighbors' NAV as soon as RTS is received
	VAR(ctrlPkt).duration = VAR(neighbNav) - VAR(durCtrlPkt);
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: Sending CTS to: %x\n", VAR(recvAddr));
#endif
	// send CTS
	TOS_CALL_COMMAND(MAC_SUB_TX_PACKET)(&VAR(ctrlPkt), sizeof(VAR(ctrlPkt)));
	VAR(state) = WAIT_DATA; // prepare to receive DATA
	VAR(radioState) = RADIO_TX;
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
}


void sendDATA() {
	// send a unicast data packet
	// assume all MAC header fields have been filled except the duration
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: Sending DATA to: %x\n", VAR(sendAddr));
#endif
#ifdef SMAC_DEBUG
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 10) = VAR(numLenErr);
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 11) = VAR(numCrcErr);
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 12) = VAR(numSlpPkt);
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 13) = VAR(numCTStimeout);
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 14) = VAR(numDATAtimeout);
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 15) = VAR(numACKtimeout);
	
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 17) = VAR(numSleeps);
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 18) = VAR(numWakeups);
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 19) = VAR(neighbNav);
#endif
	// neighbNav tracks the time I have left for tx
	VAR(dataPkt)->duration = VAR(neighbNav) - VAR(durDataPkt);
	
	TOS_CALL_COMMAND(MAC_SUB_TX_PACKET)(VAR(dataPkt), VAR(txPktLen));
	VAR(state) = WAIT_ACK;
	VAR(radioState) = RADIO_TX;
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
}


void sendACK() {
	// construct and send ACK
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: Sending ACK to: %x\n", VAR(recvAddr));
#endif
	VAR(ctrlPkt).type = ACK_PKT;
	VAR(ctrlPkt).toAddr = VAR(recvAddr);
	// stick to neighbNav -- should update it when rx data packet
	VAR(ctrlPkt).duration = VAR(neighbNav) - VAR(durCtrlPkt);
	TOS_CALL_COMMAND(MAC_SUB_TX_PACKET)(&VAR(ctrlPkt), sizeof(VAR(ctrlPkt)));
	// MAC stay in WAIT_DATA state until neighbNav becomes zero
	VAR(radioState) = RADIO_TX;
}


void sendSYNC() {
	// construct and send SYNC packet
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: Broadcasting SYNC packet: %x\n");
#endif
	VAR(syncPkt).sleepTime = VAR(schedTab)[0].counter - VAR(durSyncPkt);
	TOS_CALL_COMMAND(MAC_SUB_TX_PACKET)(&VAR(syncPkt), sizeof(VAR(syncPkt)));
	VAR(state) = IDLE;
	VAR(radioState) = RADIO_TX;
#ifdef SP_DEBUG
	ud_byte(VAR(state));
#endif
}


void startMeasure()
{
#ifdef PERFORMANCE
	VAR(clockTime) = 0;
	VAR(numRxTicks) = 0;
	VAR(numTxTicks) = 0;
	VAR(numSlpTicks) = 0;
#endif
}

